/* Shows B-tree structures. Note: only the Root DIRBLK and
   DIRBLKs in the Main Directory Band are examined. */
'@cls'
SAY
ARG drive .       /* Drive parameter missing */
IF drive = '' THEN CALL Help
IF WordPos(drive,'? /? /H HELP A: B:') \= 0 THEN CALL Help

CALL RxFuncAdd 'ReadSect','Sector','ReadSect'
CALL RxFuncAdd 'SysLoadFuncs', 'RexxUtil', 'SysLoadFuncs'
CALL SysLoadFuncs

/* Initialise Lookup Table for later bitwise ANDing */
DO exponent = 0 TO 7
   bitValue.exponent = D2C(2**exponent)
END exponent

sectorString = ReadSect(drive, 16)   /*SuperBlock is LSN 16*/
rootFnodeLSN = Bytes2Dec(13,4)
dirBandSecs = Bytes2Dec(49,4)
dirBandStart = Bytes2Dec(53,4)
dirBandEnd = Bytes2Dec(57,4)
dirBandBitmapLSN = Bytes2Dec(61,4)
numDirBlks = dirBandSecs%4 /*Each DirBlk comprises 4 secs*/

dirBlk = 0
element. = 0    /*Initialise contents of this compound var to 0*/
dirBlkHdrLen = 20
/* Read in the root FNODE */
sectorString = ReadSect(drive, rootFnodeLSN)
rootDirBlkStart = Bytes2Dec(73,4)
i = 1  /*The ShowDirBlk() func expects i to hold the count*/
element.0 = 1
element.i.dirBlkLSN = rootDirBlkStart
SAY 'Root Directory:'
CALL Charout 'STDOUT:',rootDirBlkStart"-"rootDirBlkStart+3"  "
Call ShowDirBlk /* Present the contents of the root DIRBLK */

i = 0
element. = 0
bytes = numDirBlks%8  /*Byte maps usage of 8 DIRBLKs (32 secs)*/
IF numDirBlks//8 > 0 THEN bytes = bytes+1    /* Round up */

/* Read in Dir Band Bitmap */
sector = ReadSect(drive,dirBandBitmapLSN)
/* Reduce length to valid data */
dirBandBitmapBytes = Left(sector,bytes)

DO bitmapByte = 1 TO bytes /*Examine occupancy of Dir Band*/
   char = Substr(dirBandBitmapBytes,bitmapByte,1)
   DO bitPos = 0 TO 7
      IF BitAnd(char, bitValue.bitPos) = '0'x THEN
         DO    /* DIRBLK is occupied */
         tempLSN = dirBandStart + dirBlk*4
         /* Root DIRBLK covered earlier */
         IF tempLSN \= rootDirBlkStart THEN
            DO
            i = i + 1
            element.0 = i  /*Inc count of occupied DIRBLKs*/
            element.i.dirBlkLSN=tempLSN/*Store DIRBLK start*/
            END
         END

         /* Have we reached the Directory Band end? */
         IF dirBlk+1 < numDirBlks THEN
            dirBlk = dirBlk+1 /* Not yet */
         ELSE
            LEAVE bitmapByte  /* Exit outer DO loop */

   END bitPos
END bitmapByte

IF i = 0 THEN EXIT
/* Dir Band empty (all entries must have been in root DIRBLK) */

/* Process contents of the compound variable */
DO i = 1 TO element.0
   span = element.i.dirBlkLSN'-'element.i.dirBlkLSN+3
   CALL Charout 'STDOUT:',span'  '
   CALL ShowDirBlk   /* Present the contents of a DIRBLK */
END i

EXIT  /************************************************/


Bytes2Dec:
ARG startPos,numOfChars
RETURN C2D(Reverse(Substr(sectorString,startPos,numOfChars)))


Byte2Char:
ARG startPos
RETURN Substr(sectorString,startPos,1)


ShowDirBlk:
sectorString = ''
DO k = 0 TO 3        /* Read in the 4 secs of the DIRBLK */
   sectorString = sectorString||ReadSect(drive,element.i.dirBlkLSN+k)
END k

/* Make the 0-based NextFreeByte 1-based to
   match DIRENT offset numbering */
nextFreeByte = Bytes2Dec(5,4)+1

CALL Charout 'STDOUT:','Next Byte Free:' nextFreeByte'  '

IF BitAnd(Byte2Char(9),bitValue.0) \= '0'x THEN
   DO /*Bit0 of the Change byte indicates this block is topmost*/
   SAY 'Topmost DirBlk'
   CALL CharOut 'STDOUT:',"This directory's FNODE:"
   END
ELSE
   DO
   SAY
   CALL Charout 'STDOUT:', 'Parent DirBlk:'
   END

CALL Charout 'STDOUT:',' 'Bytes2Dec(13,4)'  '

level = 1
parent = ''
CALL Charout 'STDOUT:',,
                 ' ('||FindTheRoot(Bytes2Dec(13,4))||')  '
PARSE VALUE SysCurPos() WITH row .  /* Get current cursor pos*/
CALL SysCurPos row,60    /* Set cursor to column 60 */
SAY parent               /* parent was determined by FindTheRoot() */
SAY Copies("*",50)

filenameStr = ''
dirEntOffset = dirBlkHdrLen+1 /* Convert to 1-based */
dirEnt. = ''

DO n = 0 TO 999   /* This loop will be exited by LEAVE */
   IF dirEntOffset=nextFreeByte THEN LEAVE  /*Reached end*/

   dirEntLen = Bytes2Dec(dirEntOffset,2)
   nameLen = Bytes2Dec(dirEntOffset+30,1)

   flags = Byte2Char(dirEntOffset+2,1)
   IF BitAnd(flags,bitValue.0) \= '0'x THEN
      /* Is a ".." entry */
      dirEnt.n.special = dirEnt.n.special||"S"

   IF BitAnd(flags,bitValue.2) \= '0'x THEN
      DO    /* Has a down-pointer */
      dirEnt.n.special = dirEnt.n.special||"P"
      dirEnt.n.downPtr=Bytes2Dec(dirEntOffset+dirEntLen-4,4)
   END

   IF BitAnd(flags,bitValue.3) \= '0'x THEN /*Dummy end rec*/
      dirEnt.n.special = dirEnt.n.special||"E"
   ELSE
      DO
      /* Only bother showing a set directory attribute
         if this isn't a dummy record */
      attrib = Byte2Char(dirEntOffset+3)
      IF BitAnd(attrib,bitValue.4) \= '0'x THEN
         /* Directory attribute set */
         dirEnt.n.special = dirEnt.n.special||"D"

      dirEnt.n.fnode = Bytes2Dec(dirEntOffset+4,4)
      END

   dirEnt.n.name = Substr(sectorString,dirEntOffset+31,nameLen)
   /* Translate non-printable "parent" chars */
   IF dirEnt.n.name = '101'x THEN dirEnt.n.name = '..'
   dirEnt.n.offset = dirEntOffset      /* Store offset */
   /* Adjust for next DIRENT */
   dirEntOffset = dirEntOffset+dirEntLen
END n

DO m = 0 to n-1
   IF Pos('E',dirEnt.m.special) = 0 THEN
       /* Not a dummy end record */
      CALL Charout 'STDOUT:',Left(dirEnt.m.special,3) Format(dirEnt.m.offset,5),
         '#'||Right(0||m,2)':' Strip(Right(dirEnt.m.name,15)),
         '  FNODE:'||dirEnt.m.fnode
   ELSE
      CALL Charout 'STDOUT:',Left(dirEnt.m.special,3) Format(dirEnt.m.offset,5),
         '#'||Right(0||m,2)':'  Strip(Right(dirEnt.m.name,15))

   IF Pos('P',dirEnt.m.special) > 0 THEN
      /* This DIRENT has a down-pointer */
      SAY "  DownPtr:"dirEnt.m.downPtr
   ELSE
      SAY
END m

SAY
RETURN


FindTheRoot:
ARG fnodeLSN
secString = ReadSect(drive,fnodeLSN)
/* Is this a DIRBLK or is it an FNODE? */
IF Substr(secString,1,4) = 'AE0AE477'x THEN
   DO      /* Is a DIRBLK because of Signature dword*/
   parentLSN = C2D(Reverse(Substr(secString,13,4)))
   IF level = 1 THEN
      parent = element.i.dirBlkLSN"->"||fnodeLSN"->"
   ELSE
      parent = parent||fnodeLSN"->"

   level = level+1
   CALL FindTheRoot parentLSN /*Recurse up the FNODE chain*/
   END
ELSE
   DO     /* Is an FNODE */
   nameLen = C2D(Substr(secString,13,1))
   filename = Substr(secString,14,nameLen)
   IF Length(filename) > 15 THEN
      /* Cut back name to max of 15 chars */
      filename = Left(filename,15)

   parentLSN = C2D(Reverse(Substr(secString,29,4)))

   /* Add a root dir indicator */
   IF filename = '' THEN filename = '\'

   filename = filename "[level" level"]"

   IF level = 1 THEN
      parent = element.i.dirBlkLSN"->"||fnodeLSN
   ELSE
      parent = parent||fnodeLSN

   END

RETURN filename

Help:
SAY 'ShowBtree displays the layout of any Btree structures'
SAY 'on a HPFS partition. Due to limitations in the design'
SAY 'it will not show Btree branches that are outside the'
SAY 'Directory Band, except for the Root Directory DIRBLK'
SAY 'itself.'
SAY
SAY 'Usage Example:   ShowBtree C:'
EXIT
